import colorful
from radish.utils import console_write
from radish import custom_type


def skip_step(step, resource=None, message=None):
    if resource is None:
        resource = 'any'

    if message is None:
        message = '{} {} {}'.format(colorful.orange('Can not find'),
                                    colorful.magenta(resource),
                                    colorful.orange('defined in target terraform plan.'))
    else:
        message = colorful.orange(message)

    console_write("\t{}: {}".format(colorful.bold_purple('SKIPPING'),
                                    message.format(resource=colorful.magenta(resource)))
    )
    step.skip()

    # Skip all steps in the scenario
    for each in step.parent.all_steps:
        each.runable = False


def step_condition(step):
    current_condition = step.sentence.lower().split(" ")[0]

    # if the condition is AND then check for the first previous feature line to determine if it is a
    # GIVEN, WHEN or THEN
    if current_condition == "and":
        step_id = int(step.id)-1
        if step_id > 0:
            for parent_step in step.parent.all_steps.reverse():
                # For the steps that has lower id than ours, so the steps on the above, not below
                if parent_step.id < step_id and parent_step.context_class in ["given", "when", "then"]:
                        current_condition = parent_step.context_class

    return current_condition


def write_stdout(level, message):

    prefix = colorful.bold_yellow(u'\u229b INFO :')
    if level == 'WARNING':
        prefix = colorful.bold_red(u'\u2757 WARNING :')
        message = colorful.yellow(message)


    added_prefix = u'\n\t\t{}\t{} '.format(colorful.gray(u'\u2502'),' '*len(prefix))
    message = message.split('\n')

    print(u'\t\t\u251c\u2501\t{} {}'.format(prefix, added_prefix.join(message)).encode('utf-8'))

@custom_type("ANY", r"[\"'\.\/_\-A-Za-z0-9\s:]+")
def custom_type_any(text):
    return text.replace('"', '').replace('\'', '')

@custom_type("PROPERTY", r"[\"'\.\/_\-A-Za-z0-9:\(\)\[\]]+")
def custom_type_prop(text):
    return text.replace('"', '').replace('\'', '')

@custom_type("SECTION", r"[\"'a-z]+")
def custom_type_section(text):
    if text in ['resource', 'provider', 'data', 'variable',
                'resources', 'providers', 'datas', 'variables']:
        return text.replace('"', '').replace('\'', '')

@custom_type("CONDITION", r"[\"'a-z]+")
def custom_type_condition(text):
    if text in ['only', 'not']:
        return text.replace('"', '').replace('\'', '')



from unittest import TestCase
from terraform_compliance.common.helper import (
    flatten_list,
    check_if_cidr,
    is_ip_in_cidr,
    assign_sg_params,
    validate_sg_rule,
    seek_key_in_dict,
    find_root_by_key,
    are_networks_same,
    convert_resource_type,
    seek_regex_key_in_dict_values,
    jsonify
)
from terraform_compliance.common.exceptions import Failure
from tests.mocks import MockedData


class TestHelperFunctions(TestCase):

    def test_flatten_single_dimensional_list(self):
        a = ['a', 'b', 'c']
        b = ['a', 'b', 'c']

        self.assertEqual(flatten_list(a), b)

    def test_flatten_multi_dimensional_list(self):
        a = ['a', 'b', ['c']]
        b = ['a', 'b', 'c']

        self.assertEqual(flatten_list(a), b)

    def test_flatten_multi_dimensional_nested_list(self):
        a = ['a', 'b', ['c', ['d', 'e'], 'f'], 'g', 'h', 'i', ['j', 'k']]
        b = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k']

        self.assertEqual(flatten_list(a), b)

    def test_check_if_cidr_success(self):
        self.assertTrue(check_if_cidr('10.0.0.0/8'))
        self.assertTrue(check_if_cidr('10.14.0.0/16'))
        self.assertTrue(check_if_cidr('10.0.0.0/24'))
        self.assertTrue(check_if_cidr('10.0.0.7/32'))

    def test_check_if_cidr_failure(self):
        self.assertFalse(check_if_cidr('256.0.0.0/8'))
        self.assertFalse(check_if_cidr('10.256.0.0/16'))
        self.assertFalse(check_if_cidr('10.0.256.0/24'))
        self.assertFalse(check_if_cidr('10.0.0.256/32'))
        self.assertFalse(check_if_cidr('10.0.0.256/33'))

    def test_is_ip_in_cidr_success(self):
        self.assertTrue(is_ip_in_cidr('10.0.0.0/8', ['0.0.0.0/0']))
        self.assertTrue(is_ip_in_cidr('10.0.0.0/16', ['10.0.0.0/8']))
        self.assertTrue(is_ip_in_cidr('10.0.200.0/24', ['10.0.0.0/16']))
        self.assertTrue(is_ip_in_cidr('10.0.0.1/32', ['10.0.0.0/24']))

    def test_is_ip_in_cidr_failure(self):
        self.assertFalse(is_ip_in_cidr('200.0.0.0/16', ['10.0.0.0/8']))
        self.assertFalse(is_ip_in_cidr('10.200.0.0/24', ['10.0.0.0/16']))
        self.assertFalse(is_ip_in_cidr('10.0.1.1/32', ['10.0.0.0/24']))

    def test_assign_sg_params_one_port_with_two_cidrs(self):
        self.assertEqual(MockedData.sg_params_ssh_with_2_cidrs, assign_sg_params(MockedData.sg_ssh_with_2_cidrs))

    def test_assign_sg_params_one_port_two_cidrs_any_proto(self):
        self.assertEqual(MockedData.sg_params_ssh_with_2_cidrs_any_proto, assign_sg_params(MockedData.sg_ssh_with_2_cidrs_any_proto))

    def test_assign_sg_params_all_ports_with_all_ips(self):
        self.assertEqual(MockedData.sg_params_all_port_all_ip, assign_sg_params(MockedData.sg_all_port_all_ip))

    def test_assign_sg_params_no_data_given_in_rules(self):
        self.assertEqual(MockedData.sg_params_all_port_no_ip, assign_sg_params(MockedData.sg_all_port_no_ip))

    def test_assign_sg_params_from_port_bigger_than_to_port(self):
        with self.assertRaises(Failure) as context:
            assign_sg_params(MockedData.sg_invalid)

            self.assertTrue('Invalid configuration from_port can not be bigger than to_port.' in context.exception)

    def test_validate_sg_rule_port_found_in_cidr(self):
        with self.assertRaises(Failure) as context:
            params = dict(from_port=22, to_port=22, cidr='0.0.0.0/0', ports='', proto='tcp')
            validate_sg_rule(MockedData.sg_params_all_port_all_ip, params, False)
            self.assertTrue('Found' in context.exception)

    def test_validate_sg_rule_invalid_port_range_within_scenario(self):
        with self.assertRaises(AssertionError) as context:
            params = dict(from_port=43, to_port=42, cidr=None, ports='', proto='tcp')
            validate_sg_rule(None, params, False)

            self.assertTrue('Port range is defined incorrectly within the Scenario.' in context.exception)

    def test_validate_sg_rule_port_range_found_in_cidr_fail(self):
        scenario_list = ['22-80', '21-22', '21-23', '70-72', '79-80', '79-81']
        for scenario in scenario_list:
            with self.assertRaises(Failure) as context:
                from_port, to_port = scenario.split('-')
                params = dict(proto='tcp', from_port=from_port, to_port=to_port, cidr='0.0.0.0/0', ports='')
                validate_sg_rule(MockedData.sg_params_list_range_public, params, False)
                self.assertTrue('Found' in context.exception)

    def test_validate_sg_rule_port_range_found_in_cidr_success_due_to_cidr_mismatch(self):
        scenario_list = ['22-80', '21-22', '21-23', '70-72', '79-80', '79-81']
        for scenario in scenario_list:
            from_port, to_port = scenario.split('-')
            params = dict(proto='tcp', from_port=from_port, to_port=to_port, ports='', cidr='0.0.0.0/0')
            self.assertTrue(validate_sg_rule(MockedData.sg_params_list_range_private, params, False))

    def test_validate_sg_rule_port_not_found_in_comma_delimited_scenario(self):
        with self.assertRaises(Failure) as context:
            ports = '22,443'.split(',')
            params = dict(proto='tcp', from_port=0, to_port=0, ports=ports, cidr='0.0.0.0/0')
            self.assertFalse(validate_sg_rule(MockedData.sg_params_list_range_public, params, True))

    def test_validate_sg_rule_port_found_in_comma_delimited_scenario(self):
        with self.assertRaises(Failure) as context:
            ports = range(22,80)
            ports = [str(i) for i in ports]
            params = dict(proto='tcp', from_port=0, to_port=0, ports=ports, cidr='0.0.0.0/0')
            self.assertFalse(validate_sg_rule(MockedData.sg_params_list_range_public, params, True))

    def test_seek_in_dict_finding_a_key_in_root(self):
        dictionary = dict(search_key=dict(something=[]))
        search_key = 'search_key'
        expected = [{'search_key': {'something': []}}]

        self.assertEqual(seek_key_in_dict(dictionary, search_key), expected)

    def test_seek_in_dict_finding_a_key_in_nested_dict(self):
        dictionary = dict(search_key=dict(something=dict(something_else=None)))
        search_key = 'something'
        expected = [{'something': {'something_else': None}}]

        self.assertEqual(seek_key_in_dict(dictionary, search_key), expected)

    def test_seek_in_dict_finding_multiple_keys_in_nested_dict(self):
        dictionary = dict(search_key=dict(something=dict(something_else=None, something=['something_else'])), something=[])
        search_key = 'something'
        expected = [{'something': {'something_else': None, 'something': ['something_else']}}, {'something': []}]

        self.assertEqual(seek_key_in_dict(dictionary, search_key), expected)

    def test_seek_in_dict_finding_values_in_non_dicts(self):
        dictionary = dict(search_key=dict(something=[]))
        search_key = 'something_else'
        expected = []

        self.assertEqual(seek_key_in_dict(dictionary, search_key), expected)

    def test_seek_in_dict_finding_values_in_non_dicts_on_root(self):
        dictionary = 'something_else'
        search_key = 'something_else'
        expected = []

        self.assertEqual(seek_key_in_dict(dictionary, search_key), expected)

    def test_find_root_by_key_single_return(self):
        haystack = dict(some_key=dict(values=dict(tags=[], something_else='something')))
        search_key = 'tags'
        expected = ['some_key']

        self.assertEqual(find_root_by_key(haystack, search_key), expected)

    def test_find_root_by_key_multiple_return(self):
        haystack = dict(some_key=dict(values=dict(tags=[], something_else='something')), other_key=dict(values=dict(tags=[], something_else='something')))
        search_key = 'tags'
        expected = ['other_key', 'some_key']

        self.assertEqual(sorted(find_root_by_key(haystack, search_key)), sorted(expected))

    def test_find_root_by_key_multiple_return_02(self):
        haystack = dict(some_key=dict(values=dict(tags=[], something_else='loki'), find_me='bingo'),
                        other_key=dict(values=dict(tags=[], something_else='thor'), find_me='b i n g o'))
        search_key = 'tags'
        return_key = 'find_me'
        expected = ['bingo', 'b i n g o']

        self.assertEqual(sorted(find_root_by_key(haystack, search_key, return_key)), sorted(expected))

    def test_are_networks_same_success(self):
        network_a = '192.168.0.0/24'
        networks = ['192.168.0.0/24']
        self.assertTrue(are_networks_same(network_a, networks))

    def test_are_networks_same_false(self):
        network_a = '192.168.0.0/24'
        networks = ['192.168.0.0/23']
        self.assertFalse(are_networks_same(network_a, networks))

    def test_convert_resource_type_success(self):
        self.assertEqual(convert_resource_type('AWS Security Group'), 'aws_security_group')

    def test_convert_resource_type_failure(self):
        self.assertEqual(convert_resource_type('test_resource'), 'test_resource')

    def test_seek_regex_in_dict_value_nested_dict(self):
        haystack = dict(search_key=dict(something='value'))
        key_name = 'something'
        needle = 'val.*'
        expected = ['value']

        self.assertEqual(seek_regex_key_in_dict_values(haystack, key_name, needle), expected)

    def test_jsonify(self):
        self.assertEqual(jsonify({}), {})
        self.assertEqual(jsonify([]), [])
        self.assertEqual(jsonify(12), 12)
        self.assertEqual(jsonify('something'), 'something')
        self.assertEqual(jsonify('{"test": "test_value"}'), {'test': 'test_value'})



import sys
import os
from argparse import Action


class ReadableDir(Action):
    def __init__(self, option_strings, dest, **kwargs):
        super(ReadableDir, self).__init__(option_strings, dest, **kwargs)
        self.option_strings = option_strings
        self.dest = dest
        self.nargs = kwargs.get('nargs', None)
        self.const = kwargs.get('const', None)
        self.default = kwargs.get('default', None)
        self.type = kwargs.get('type', None)
        self.choices = kwargs.get('choices', None)
        self.required = kwargs.get('required', None)
        self.help = kwargs.get('help', None)
        self.metavar = kwargs.get('metavar', None)

    def __call__(self, parser, namespace, values, option_string=None):
        prospective_dir = values

        # Check if the given directory is actually a git repo
        if prospective_dir.startswith('git:'):
            print('Using remote git repository: {}'.format(prospective_dir[4:]))
            setattr(namespace, self.dest, prospective_dir[4:])
            return True

        # Check if the given path is a directory really
        if not os.path.isdir(prospective_dir):
            print('ERROR: {} is not a directory.'.format(prospective_dir))
            sys.exit(1)

        # Check if we have access to that directory
        if os.access(prospective_dir, os.R_OK):
            setattr(namespace, self.dest, prospective_dir)
            return True

        print('ERROR: Can not read {}'.format(prospective_dir))
        sys.exit(1)



from unittest import TestCase
from terraform_compliance.common.readable_plan import ReadablePlan, filetype
from terraform_compliance.common.exceptions import TerraformComplianceInternalFailure

import os
from mock import patch


class TestReadablePlan(TestCase):

    @patch.object(os.path, 'isfile', return_value=False)
    def test_readable_plan_is_not_a_file(self, *args):
        class Namespace(object):
            pass

        with self.assertRaises(SystemExit):
            ReadablePlan('parser', 'value').__call__('parser', Namespace, 'not_a_file')

    @patch.object(os.path, 'isfile', return_value=True)
    @patch.object(filetype, 'guess', return_value='something')
    @patch.object(os.path, 'isfile', return_value=True)
    @patch.object(filetype, 'guess', return_value=None)
    @patch('terraform_compliance.common.readable_plan.convert_terraform_plan_to_json', return_value=None)
    def test_readable_plan_filetype_guess_is_none(self, *args):
        class Namespace(object):
            pass

        with self.assertRaises(TerraformComplianceInternalFailure):
            ReadablePlan('parser', 'value').__call__('parser', Namespace, 'git:value')

    @patch.object(os.path, 'isfile', return_value=True)
    @patch.object(filetype, 'guess', return_value='something')
    @patch('terraform_compliance.common.readable_plan.convert_terraform_plan_to_json', return_value='something')

    def test_readable_plan_convert_terraform_files(self, *args):
        class Namespace(object):
            pass

        with self.assertRaises(TerraformComplianceInternalFailure):
            ReadablePlan('parser', 'value').__call__('parser', Namespace, 'git:value')

    @patch.object(os.path, 'isfile', return_value=True)
    @patch.object(filetype, 'guess', return_value=None)
    def test_readable_plan_filetype_guess_is_none(self, *args):
        class Namespace(object):
            pass

        with self.assertRaises(TerraformComplianceInternalFailure):
            ReadablePlan('parser', 'value').__call__('parser', Namespace, 'git:value')

    @patch.object(os.path, 'isfile', return_value=True)
    @patch.object(filetype, 'guess', return_value=True)
    @patch('terraform_compliance.common.readable_plan.convert_terraform_plan_to_json', return_value='/bin/ls')
    def test_readable_plan_file_is_not_json(self, *args):
        class Namespace(object):
            pass

        with self.assertRaises(SystemExit):
            ReadablePlan('parser', 'value').__call__('parser', Namespace, 'git:value')



import re
from netaddr import IPNetwork
from terraform_compliance.steps import resource_name
from terraform_compliance.common.exceptions import Failure
from collections.abc import Iterable
import json


class EmptyStash(object):
    pass


class Null(object):
    pass


def flatten_list(input):
    return list(flatten(input))


def flatten(items):
    for x in items:
        if isinstance(x, Iterable) and not isinstance(x, (str, bytes)):
            for sub_x in flatten(x):
                yield sub_x
        else:
            yield x


def check_if_cidr( value ):
    regex = r'(1[0-9][0-9]|2[0-4][0-9]|25[0-5]|[0-9][0-9]|[0-9])\.(1[0-9][0-9]|2[0-4][0-9]|25[0-5]|[0-9][0-9]|[0-9])\.(1[0-9][0-9]|2[0-4][0-9]|25[0-5]|[0-9][0-9]|[0-9])\.(1[0-9][0-9]|2[0-4][0-9]|25[0-5]|[0-9][0-9]|[0-9])\/(3[0-2]|2[0-9]|1[0-9]|[0-9])'
    matches = re.match(regex, value)

    if matches is not None:
        return True

    return False


def is_ip_in_cidr(ip_cidr, cidr):
    for ip_network in cidr:
        if check_if_cidr(ip_cidr) and check_if_cidr(ip_network) and IPNetwork(ip_cidr) in IPNetwork(ip_network):
            return True

    return False


def are_networks_same(first_network, network_list):
    for second_network in network_list:
        if check_if_cidr(first_network) and check_if_cidr(second_network) and \
                IPNetwork(first_network) == IPNetwork(second_network):
            return True

    return False


# A helper function that compares port related data with given dictionary
def check_sg_rules(plan_data, security_group, condition):
    return validate_sg_rule(plan_data=assign_sg_params(plan_data),
                            params=security_group,
                            condition=condition)


def assign_sg_params(rule):
    from_port = int(rule.get('from_port', 0))
    to_port = int(rule.get('to_port', 0))

    protocol = [proto for proto in [rule.get('protocol', '-1')]]

    if protocol[0] == '-1' or type(protocol[0]) is int:
        protocol = ['tcp', 'udp']

    protocol[0] = protocol[0].lower()

    cidr_blocks = rule.get('cidr_blocks', [])

    if type(cidr_blocks) is not list:
        cidr_blocks = [cidr_blocks]

    if to_port == 0 and from_port == 0:
        to_port = 65535

    if from_port > to_port:
        raise Failure('Invalid configuration from_port can not be bigger than to_port. '
                                         '{} > {} {} in {}'.format(from_port,
                                                                   to_port,
                                                                   protocol,
                                                                   cidr_blocks))

    return dict(protocol=protocol, from_port=from_port, to_port=to_port, cidr_blocks=cidr_blocks)


def validate_sg_rule(plan_data, params, condition):
    from_port = int(params['from_port'])
    to_port = int(params['to_port'])

    assert from_port <= to_port, 'Port range is defined incorrectly within the Scenario. ' \
                                'Define it {}-{} instead of {}-{}.'.format(from_port,
                                                                           to_port,
                                                                           to_port,
                                                                           from_port)

    defined_range = set(range(plan_data['from_port'], plan_data['to_port']+1))
    defined_network_list = plan_data['cidr_blocks']
    given_network = params.get('cidr', None)

    # Condition: must only have
    # Fail only if ;
    # * the ports does not match and defined network is a subset of given network.
    if condition:
        given_range = set([int(port) for port in params['ports']])
        from_to_port = ','.join(params['ports'])

        # Set to True if ports are exactly same.
        port_intersection = given_range == defined_range

        # Set to True if one of the given networks is a subset.
        network_check = is_ip_in_cidr(given_network, defined_network_list)

        if not port_intersection and network_check:
            raise Failure('{}/{} ports are defined for {} network. '
                          'Must be limited to {}/{} and {}'.format('/'.join(plan_data['protocol']),
                                                            '{}-{}'.format(plan_data['from_port'],
                                                                           plan_data['to_port']),
                                                            plan_data['cidr_blocks'],
                                                            '/'.join(plan_data['protocol']),
                                                            from_to_port,
                                                            given_network))

    # Condition: must not have
    # Fail only if ;
    # * the ports match and networks match
    else:
        given_range = set(range(from_port, to_port+1))
        port_intersection = given_range & defined_range
        from_to_port = str(from_port) + '-' + str(to_port)
        network_intersection = is_ip_in_cidr(given_network, defined_network_list)

        if port_intersection and network_intersection:
            raise Failure('{}/{} ports are defined for {} network.'.format('/'.join(plan_data['protocol']),
                                                                           '{}-{}'.format(plan_data['from_port'],
                                                                                          plan_data['to_port']),
                                                                           plan_data['cidr_blocks']))

    return True


def convert_resource_type(resource_type):
    '''
    Searchs given resource_type within resource_name array and returns the value if it is found

    :param resource_type: String of resource_type
    :return: converted or original resource_type
    '''
    if resource_type in resource_name.keys():
        resource_type = resource_name[resource_type]

    return resource_type


def seek_key_in_dict(haystack, needle):
    '''
    Searches needle in haystack ( could be dict, list, list of dicts, nested dicts, etc. ) and returns all findings
    as a list

    :param haystack: dict, list
    :param needle: search key
    :return: list of found keys & values
    '''
    found = list()
    if type(haystack) is dict:
        for key, value in haystack.items():
            if key.lower() == needle.lower():
                found.append({key: value})
            else:
                found.extend(seek_key_in_dict(value, needle))

    elif type(haystack) is list:
        for value in haystack:
            found.extend(seek_key_in_dict(value, needle))

    else:
        return []

    return found


def seek_regex_key_in_dict_values(haystack, key_name, needle, key_matched=None):
    '''
    Searches needle in haystack ( could be dict, list, list of dicts, nested dicts, etc. ) and returns all findings
    as a list. The only difference from seek_key_in_dict is, we are assuming needle is in regex format here and we
    are searching for values instead.

    :param haystack: dict, list
    :param key_name: string of the key
    :param needle: regex search for the value
    :param key_matched: Internal use
    :return: list of found keys & values
    '''
    regex = r'{}'.format(needle)
    found = list()
    if type(haystack) is dict:
        for key, value in haystack.items():
            if key.lower() == key_name.lower() or key_matched is not None:
                if type(value) is str:
                    matches = re.match(regex, value)

                    if matches is not None:
                        found.append(matches.group(0))
                    else:
                        found.extend(seek_regex_key_in_dict_values(value, key_name, needle, True))

                elif type(value) is dict:
                    found.extend(seek_regex_key_in_dict_values(value, key_name, needle, True))

                elif type(value) is list:
                    for value in haystack:
                        found.extend(seek_regex_key_in_dict_values(value, key_name, needle, True))

            else:
                found.extend(seek_regex_key_in_dict_values(value, key_name, needle, key_matched))

    elif type(haystack) is list:
        for value in haystack:
            found.extend(seek_regex_key_in_dict_values(value, key_name, needle, key_matched))

    else:
        return []

    return found


def find_root_by_key(haystack, needle, return_key=None, _inherited_key=None, _depth=0, _return_value=None):
    '''
    Searches needle in haystack ( could be dict, list, list of dicts, nested dicts, etc. ) and returns the root key
    that has this needle somewhere within it's children.

    :param haystack: dict, list
    :param needle: search key
    :param return_key: if this is given, then the result will be the root_key[return_key] instead of root_key
    :param _inherited_key: internal usage, do not pass this.
    :param _depth: internal usage, do not pass this.
    :param _return_value: internal usage, do not pass this.
    :return:
    '''
    found = list()
    if type(haystack) is dict:
        for key, value in haystack.items():
            if not _depth:
                _inherited_key = key
                _return_value = key if not return_key else haystack[_inherited_key].get(return_key, _inherited_key)

            if key.lower() == needle.lower():
                found.append(_return_value)
            else:
                found.extend(find_root_by_key(value, needle, return_key, _inherited_key, _depth+1, _return_value))

    elif type(haystack) is list and _inherited_key is not None:
        for value in haystack:
            found.extend(find_root_by_key(value, needle, return_key, _inherited_key, _depth+1, _return_value))

    else:
        return []

    return list(set(found))


def jsonify(string):
    if type(string) is not str:
        return string

    try:
        return json.loads(string)
    except json.decoder.JSONDecodeError:
        return string


def get_resource_name_from_stash(stash, alternative_stash=None):
    if type(alternative_stash) is str or type(alternative_stash) is bool or alternative_stash is None:
        if type(stash) is list:

            # Get the first number, since this is usually due to `count` usage in terraform
            if 'address' in stash[0]:
                return {'address': stash[0]['address'].replace('[0]','')}
            else:
                return {'address': stash[0]}

        else:
            if 'address' in alternative_stash:
                return alternative_stash
            else:
                return {'address': alternative_stash}



# -*- coding: utf-8 -*-

from radish import world, given, when, then, step
from terraform_compliance.steps import property_match_list
from terraform_compliance.common.helper import check_sg_rules, convert_resource_type, find_root_by_key, seek_key_in_dict
from terraform_compliance.common.helper import seek_regex_key_in_dict_values, jsonify, Null, EmptyStash
from terraform_compliance.common.helper import get_resource_name_from_stash
from terraform_compliance.extensions.ext_radish_bdd import skip_step
from terraform_compliance.extensions.ext_radish_bdd import custom_type_any
from terraform_compliance.extensions.ext_radish_bdd import custom_type_condition
from terraform_compliance.extensions.ext_radish_bdd import custom_type_section
from terraform_compliance.extensions.ext_radish_bdd import custom_type_prop
import re
from terraform_compliance.common.exceptions import Failure, TerraformComplianceNotImplemented
from terraform_compliance.common.exceptions import TerraformComplianceInternalFailure


# TODO: Figure out how the IAM policies/statements shown in the plan.out
# TODO: Implement an IAM Compliance via https://github.com/Netflix-Skunkworks/policyuniverse



# def i have_name_defined(_step_obj, name, _terraform_config=world):
#     pass

@given(u'I have {name:ANY} defined')
@given(u'I have {name:ANY} {type_name:SECTION} configured')
def i_have_name_section_configured(_step_obj, name, type_name='resource', _terraform_config=world):
    '''
    Finds given resource or variable by name and returns it. Skips the step (and further steps) if it is not found.

    :param _step_obj: Internal, step object for radish.
    :param name: String of the name of the resource_type or variable.
    :param type_name: String of the type, either resource(s) or variable(s)
    :param _terraform_config: Internal, terraform configuration.
    :return:
    '''
    assert (type_name in ['resource', 'resources',
                          'variable', 'variables',
                          'provider', 'providers',
                          'data', 'datas']), \
        '{} configuration type does not exist or not implemented yet. ' \
        'Use resource(s), provider(s), variable(s) or data(s) instead.'.format(type_name)

    if type_name.endswith('s'):
        type_name = type_name[:-1]

    if name == 'resource that supports tags':
        resource_types_supports_tags = find_root_by_key(_terraform_config.config.terraform.resources_raw,
                                                        'tags',
                                                        return_key='type')
        resource_list = []
        for resource_type in resource_types_supports_tags:
            resource_list.extend(_terraform_config.config.terraform.find_resources_by_type(resource_type))

        if resource_list:
            _step_obj.context.type = type_name
            _step_obj.context.name = name
            _step_obj.context.stash = resource_list
            return True

    elif type_name == 'resource':
        name = convert_resource_type(name)
        resource_list = _terraform_config.config.terraform.find_resources_by_type(name)

        if resource_list:
            _step_obj.context.type = type_name
            _step_obj.context.name = name
            _step_obj.context.stash = resource_list
            return True

    elif type_name == 'variable':
        found_variable = _terraform_config.config.terraform.variables.get(name, None)

        if found_variable:
            _step_obj.context.type = type_name
            _step_obj.context.name = name
            _step_obj.context.stash = found_variable
            return True

    elif type_name == 'provider':
        found_provider = _terraform_config.config.terraform.configuration.get('providers', {}).get(name, None)

        if found_provider:
            _step_obj.context.type = type_name
            _step_obj.context.name = name
            _step_obj.context.stash = found_provider
            return True

    elif type_name == 'data':
        name = convert_resource_type(name)
        data_list = _terraform_config.config.terraform.find_data_by_type(name)

        if data_list:
            _step_obj.context.type = type_name
            _step_obj.context.name = name
            _step_obj.context.stash = data_list
            return True

    skip_step(_step_obj, name)


@when(u'its {key:PROPERTY} is {value:PROPERTY}')
@when(u'its {key:PROPERTY} has {value:PROPERTY}')
@when(u'its {key:PROPERTY} includes {value:PROPERTY}')
@when(u'its {key:PROPERTY} contains {value:PROPERTY}')
def its_key_is_value(_step_obj, key, value):
    search_key = str(key).lower()
    found_list = []
    for obj in _step_obj.context.stash:
        object_key = obj.get(key, Null)

        if object_key is Null:
            object_key = obj.get('values', {})
            if type(object_key) is list:
                object_keys = []
                for object_key_element in object_key:
                    object_keys.append(object_key_element.get(key, Null))

                object_key = [keys for keys in object_keys if keys is not Null]
            else:
                object_key = object_key.get(key, Null)

        if type(object_key) is str:
            if "[" in object_key:
                object_key = object_key.split('[')[0]

            if object_key == value:
                found_list.append(obj)

        elif type(object_key) in (int, bool) and object_key == value:
            found_list.append(obj)

        elif type(object_key) is list and value in object_key:
            found_list.append(obj)

        elif type(object_key) is dict and (value in object_key.keys()):
            found_list.append(obj)

    if found_list is not []:
        _step_obj.context.stash = found_list
    else:
        skip_step(_step_obj, value)

@when(u'its {key:PROPERTY} is not {value:PROPERTY}')
@when(u'its {key:PROPERTY} has not {value:PROPERTY}')
@when(u'its {key:PROPERTY} does not include {value:PROPERTY}')
@when(u'its {key:PROPERTY} does not contain {value:PROPERTY}')
def its_key_is_not_value(_step_obj, key, value):
    search_key = str(key).lower()
    found_list = []
    for obj in _step_obj.context.stash:
        object_key = obj.get(key, Null)

        if object_key is Null:
            object_key = obj.get('values', {})
            if type(object_key) is list:
                object_keys = []
                for object_key_element in object_key:
                    if object_key_element.get(key, Null) != value:
                        object_keys.append(object_key_element.get(key, Null))

                object_key = [keys for keys in object_keys if keys is not Null]
            else:
                if object_key.get(key, Null) != value:
                    object_key = object_key.get(key, Null)

        if type(object_key) is str:
            if "[" in object_key:
                object_key = object_key.split('[')[0]

            if object_key != value:
                found_list.append(obj)

        elif type(object_key) in (int, bool) and object_key != value:
            found_list.append(obj)

        elif type(object_key) is list and value not in object_key:
            found_list.append(obj)

        elif type(object_key) is dict and (value in object_key.keys()):
            found_list.append(obj)

    if found_list is not []:
        _step_obj.context.stash = found_list
    else:
        skip_step(_step_obj, value)

@when(u'it contain {something:ANY}')
@when(u'they have {something:ANY}')
@when(u'it has {something:ANY}')
@when(u'it contains {something:ANY}')
@then(u'it must contain {something:ANY}')
def it_condition_contain_something(_step_obj, something):
    prop_list = []

    if _step_obj.context.type in ('resource', 'data'):
        for resource in _step_obj.context.stash:
            if type(resource) is not dict:
                resource = {'values': resource,
                            'address': resource,
                            'type': _step_obj.context.name}

            values = resource.get('values', resource.get('expressions', {}))

            found_value = Null
            found_key = Null
            if type(values) is dict:
                found_key = values.get(something, seek_key_in_dict(values, something))
                if type(found_key) is not list:
                    found_key = [{something: found_key}]

                if len(found_key):
                    found_key = found_key[0] if len(found_key) == 1 else found_key

                    if type(found_key) is dict:
                        found_value = jsonify(found_key.get(something, found_key))
                    else:
                        found_value = found_key
            elif type(values) is list:
                found_value = []

                for value in values:

                    if type(value) is dict:
                        # First search in the keys
                        found_key = seek_key_in_dict(value, something)

                        # Then search in the values with 'key'
                        if not found_key:
                            found_key = seek_regex_key_in_dict_values(value, 'key', something)

                            if found_key:
                                found_key = found_key[0]
                                found_value = value.get('value')
                                break

                    if found_key is not Null and len(found_key):
                        found_key = found_key[0] if len(found_key) == 1 else found_key

                        if type(found_key) is dict:
                            found_value.append(jsonify(found_key.get(something, found_key)))

            if type(found_value) is dict and 'constant_value' in found_value:
                found_value = found_value['constant_value']

            if found_value is not Null and found_value != [] and found_value != '' and found_value != {}:
                prop_list.append({'address': resource['address'],
                                  'values': found_value,
                                  'type': _step_obj.context.name})

            elif 'must' in _step_obj.context_sensitive_sentence:
                raise Failure('{} ({}) does not have {} property.'.format(resource['address'],
                                                                          resource.get('type', ''),
                                                                          something))

        if prop_list:
            _step_obj.context.stash = prop_list
            _step_obj.context.property_name = something
            return True

        skip_step(_step_obj,
                  resource=_step_obj.context.name,
                  message='Can not find any {} property for {} resource in '
                          'terraform plan.'.format(something, _step_obj.context.name))

    elif _step_obj.context.type == 'provider':
        values = seek_key_in_dict(_step_obj.context.stash, something)

        if values:
            _step_obj.context.stash = values
            _step_obj.context.property_name = something
            return True

    skip_step(_step_obj,
              resource=_step_obj.context.name,
              message='Skipping the step since {} type does not have {} property.'.format(_step_obj.context.type,
                                                                                          something))



@then(u'{something:ANY} is be enabled')
@then(u'{something:ANY} must be enabled')
def property_is_enabled(_step_obj, something):
    for resource in _step_obj.context.stash:
        if type(resource) is dict:
            if something in property_match_list:
                something = property_match_list[something].get(resource['type'], something)

            property_value = seek_key_in_dict(resource.get('values', {}), something)

            if len(property_value):
                property_value = property_value[0]

                if type(property_value) is dict:
                    property_value = property_value.get(something, Null)

            if not property_value:
                raise Failure('Resource {} does not have {} property enabled ({}={}).'.format(resource.get('address', "resource"),
                                                                                              something,
                                                                                              something,
                                                                                              property_value))
    return True


@then(u'it must {condition:ANY} have {proto:ANY} protocol and port {port} for {cidr:ANY}')
def it_condition_have_proto_protocol_and_port_port_for_cidr(_step_obj, condition, proto, port, cidr):
    proto = str(proto)
    cidr = str(cidr)

    # Set to True only if the condition is 'only'
    condition = condition == 'only'

    # In case we have a range
    if '-' in port:
        if condition:
            raise Failure('"must only" scenario cases must be used either with individual port '
                          'or multiple ports separated with comma.')

        from_port, to_port = port.split('-')
        ports = [from_port, to_port]

    # In case we have comma delimited ports
    elif ',' in port:
        ports = [port for port in port.split(',')]
        from_port = min(ports)
        to_port = max(ports)

    else:
        from_port = to_port = int(port)
        ports = list(set([str(from_port), str(to_port)]))

    from_port = int(from_port) if int(from_port) > 0 else 1
    to_port = int(to_port) if int(to_port) > 0 else 1
    ports[0] = ports[0] if int(ports[0]) > 0 else '1'

    looking_for = dict(proto=proto,
                       from_port=int(from_port),
                       to_port=int(to_port),
                       ports=ports,
                       cidr=cidr)

    for security_group in _step_obj.context.stash:
        if type(security_group['values']) is list:
            for sg in security_group['values']:
                check_sg_rules(plan_data=sg, security_group=looking_for, condition=condition)

        elif type(security_group['values']) is dict:
            check_sg_rules(plan_data=security_group['values'], security_group=looking_for, condition=condition)
        else:
            raise TerraformComplianceInternalFailure('Unexpected Security Group, '
                                                     'must be either list or a dict: '
                                                     '{}'.format(security_group['values']))
    return True

@when(u'I {action_type:ANY} it')
@when(u'I {action_type:ANY} them')
@when(u'I {action_type:ANY} the value')
@then(u'I {action_type:ANY} it')
@then(u'I {action_type:ANY} them')
@then(u'I {action_type:ANY} the value')
def i_action_them(_step_obj, action_type):
    if action_type == "count":
        # WARNING: Only case where we set stash as a dictionary, instead of a list.
        if type(_step_obj.context.stash) is list:
            if type(_step_obj.context.stash[0]) is dict():
                if _step_obj.context.stash.get('values'):
                    _step_obj.context.stash = seek_key_in_dict(_step_obj.context.stash, 'values')
                    count = 0
                    for result in _step_obj.context.stash:
                        count += len(result.get('values', {})) if result.get('values') else 1

                    _step_obj.context.stash = {'values': count}
            else:
                _step_obj.context.stash = {'values': len(_step_obj.context.stash)}
    else:
        raise TerraformComplianceNotImplemented('Invalid action_type in the scenario: {}'.format(action_type))


@then(u'Its value must be {operator:ANY} than {number:d}')
@then(u'I expect the result is {operator:ANY} than {number:d}')
@then(u'Its value must be {operator:ANY} to {number:d}')
@then(u'I expect the result is {operator:ANY} to {number:d}')
def i_expect_the_result_is_operator_than_number(_step_obj, operator, number, _stash=EmptyStash):
    values = _step_obj.context.stash if _stash is EmptyStash else _stash

    if type(values) is list:
        for value_set in values:
            i_expect_the_result_is_operator_than_number(_step_obj, operator, number, _stash=value_set)

    elif type(values) is dict:
        i_expect_the_result_is_operator_than_number(_step_obj, operator, number, values.get('values', Null))

    elif type(values) is int or type(values) is str:
        values = int(values)
        if operator in ("more", "greater", "bigger"):
            assert values > number, "{} is not more than {}".format(values, number)
        elif operator in ("more and equal", "greater and equal", "bigger and equal"):
            assert values >= number, "{} is not more and equal than {}".format(values, number)
        elif operator in ("less", "lesser", "smaller"):
            assert values < number, "{} is not less than {}".format(values, number)
        elif operator in ("less and equal", "lesser and equal", "smaller and equal"):
            assert values <= number, "{} is not less and equal than {}".format(values, number)
        elif operator in ("equal",):
            assert values == number, "{} is not equal to {}".format(values, number)
        else:
            raise TerraformComplianceNotImplemented('Invalid operator: {}'.format(operator))

    elif type(values) is Null:
        raise TerraformComplianceNotImplemented('Null/Empty value found on {}'.format(_step_obj.context.type))


@step(u'its value {condition:ANY} match the "{search_regex}" regex')
def its_value_condition_match_the_search_regex_regex(_step_obj, condition, search_regex, _stash=EmptyStash):
    def fail(condition, name=None):
        text = 'matches' if condition == 'must not' else 'does not match'
        name = name if (name is not None or name is not False) else _step_obj.context.name
        pattern = 'Null/None' if regex == '\x00' else regex
        raise Failure('{} property in {} {} {} with {} regex. '
                      'It is set to {}.'.format(_step_obj.context.property_name,
                                                name,
                                                _step_obj.context.type,
                                                text,
                                                pattern,
                                                values))
    regex = r'{}'.format(search_regex)
    values = _step_obj.context.stash if _stash is EmptyStash else _stash

    if type(values) in (str, int, bool) or values is None:
        matches = re.match(regex, str(values), flags=re.IGNORECASE)

        if (condition == 'must' and matches is None) or (condition == "must not" and matches is not None):
            _stash = get_resource_name_from_stash(_step_obj.context.stash, _stash)
            fail(condition, name=_stash.get('address'))

    elif type(values) is list:
        for value in values:
            its_value_condition_match_the_search_regex_regex(_step_obj, condition, search_regex, value)

    elif type(values) is dict:
        if 'values' in values:
            if values['values'] is None and regex == '\x00' and condition == 'must not':
                values = values['values']
                fail(condition, name=_stash.get('address'))
            else:
                its_value_condition_match_the_search_regex_regex(_step_obj, condition, search_regex, values.get('values'))

        else:
            for key, value in values.items():
                its_value_condition_match_the_search_regex_regex(_step_obj, condition, search_regex, value)


@step(u'its value {condition:ANY} be {match:ANY}')
def its_value_condition_equal(_step_obj, condition, match, _stash=EmptyStash):
    its_value_condition_match_the_search_regex_regex(_step_obj, condition, "^" + re.escape(match) + "$", _stash)


@then(u'its value {condition:ANY} contain {value:ANY}')
def its_value_condition_contain(_step_obj, condition, value, _stash=EmptyStash):
    values = _step_obj.context.stash if _stash is EmptyStash else _stash
    if isinstance(values, list):
        for value_set in values:
            its_value_condition_contain(_step_obj, condition, value, value_set)
    elif isinstance(values, dict):
        _its_value_condition_contain(_step_obj, condition, value, values.get('values', Null))
    elif type(values) is Null:
        raise TerraformComplianceNotImplemented('Null/Empty value found on {}'.format(_step_obj.context.type))


def _its_value_condition_contain(_step_obj, condition, value, values):
    assert condition in ('must', 'must not'), 'Condition should be one of: `must`, `must not`'
    if isinstance(values, (int, bool, str, float)):
        values = [values]

    if isinstance(values, list):
        values = [str(v) for v in values]
        _fail_text = 'did not contain' if condition == 'must' else 'contains'
        fail_message = '{} property in {} {} {}, it is set to {}'.format(
            _step_obj.context.property_name,
            _step_obj.context.name,
            _fail_text,
            value,
            values,
        )
        if condition == 'must':
            assert value in values, fail_message
        else:
            assert value not in values, fail_message
    else:
        raise Failure('Can only check that if list contains value')


@then(u'the scenario fails')
@then(u'the scenario should fail')
@then(u'it fails')
@then(u'it should fail')
@then(u'it must fail')
def it_fails(_step_obj):
    raise Failure('Forcefully failing the scenario on {} {}'.format(_step_obj.context.name,
                                                                    _step_obj.context.type))


@then(u'its value {condition:ANY} be null')
def its_value_condition_be_null(_step_obj, condition):
    its_value_condition_match_the_search_regex_regex(_step_obj, condition, u'\x00')
    its_value_condition_match_the_search_regex_regex(_step_obj, condition, u'^$')
    its_value_condition_match_the_search_regex_regex(_step_obj, condition, u'^null$')



from unittest import TestCase
from terraform_compliance.extensions.ext_radish_bdd import skip_step, write_stdout, step_condition
from tests.mocks import MockedStep
import sys

class TestRadishBddExtension(TestCase):
    def test_step_condition(self):
        step = MockedStep()

        step.sentence = 'Given I do this'
        self.assertEqual(step_condition(step), 'given')

        step.sentence = 'When I do this'
        self.assertEqual(step_condition(step), 'when')

        step.sentence = 'Then I do this'
        self.assertEqual(step_condition(step), 'then')

    def test_write_stdout(self):
        self.assertEqual(None, write_stdout('info','test'))

    def test_skip_step(self):
        step = MockedStep()
        skip_step(step)
        self.assertEqual(step.state, 'skipped')




class TerraformComplianceInvalidConfig(Exception):
    pass

class TerraformComplianceInvalidConfigurationType(Exception):
    pass

class Failure(Exception):
    pass

class TerraformComplianceNotImplemented(Exception):
    pass

class TerraformComplianceInternalFailure(Exception):
    pass



import os
from argparse import ArgumentParser
from tempfile import mkdtemp
from git import Repo
from terraform_compliance import __app_name__, __version__
from terraform_compliance.common.readable_dir import ReadableDir
from terraform_compliance.common.readable_plan import ReadablePlan
from radish.main import main as call_radish


print('{} v{} initiated\n'.format(__app_name__, __version__))


class ArgHandling(object):
    pass


def cli(arghandling=ArgHandling(), argparser=ArgumentParser(prog=__app_name__,
                                                            description="BDD Test Framework for Hashicorp terraform")):
    args = arghandling
    parser = argparser
    parser.add_argument("--terraform", "-t", dest="terraform_file", metavar='terraform_file', type=str, nargs='?',
                        help="The absolute path to the terraform executable.", required=False)
    _, radish_arguments = parser.parse_known_args(namespace=args)

    # TODO: Create a custom usage() since it just shows -t right now

    parser.add_argument("--features", "-f", dest="features", metavar='feature directory', action=ReadableDir,
                        help="Directory (or git repository with 'git:' prefix) consists of BDD features", required=True)
    parser.add_argument("--planfile", "-p", dest="plan_file", metavar='plan_file', action=ReadablePlan,
                        help="Plan output file generated by Terraform", required=True)
    parser.add_argument("--identity", "-i", dest="ssh_key", metavar='ssh private key', type=str, nargs='?',
                        help="SSH Private key that will be use on git authentication.", required=False)
    parser.add_argument("--version", "-v", action="version", version=__version__)

    _, radish_arguments = parser.parse_known_args(namespace=args)

    steps_directory = os.path.join(os.path.split(os.path.abspath(__file__))[0], 'steps')

    # SSH Key is given for git authentication
    ssh_cmd = {}
    if args.ssh_key:
        ssh_cmd = {"GIT_SSH_COMMAND": "ssh -l {} -i {}".format('git', args.ssh_key)}

    # A remote repository used here
    if args.features.startswith(('http', 'https', 'ssh')):
        features_git_repo = args.features
        args.features = mkdtemp()

        Repo.clone_from(url=features_git_repo, to_path=args.features, env=ssh_cmd)

    features_directory = os.path.join(os.path.abspath(args.features))
    print('* Features  : {}{}'.format(features_directory,
                                     (' ({})'.format(features_git_repo) if 'features_git_repo' in locals() else '')))

    print('* Plan File : {}'.format(args.plan_file))
    commands = ['radish',
                '--write-steps-once',
                features_directory,
                '--basedir', steps_directory,
                '--user-data=plan_file={}'.format(args.plan_file)]
    commands.extend(radish_arguments)

    print('\n. Running tests.')
    result = call_radish(args=commands[1:])

    return result


if __name__ == '__main__':
    cli()



from radish import before, world
from terraform_compliance.extensions.terraform import TerraformParser


@before.each_feature
def load_terraform_data(feature):
    world.config.terraform = TerraformParser(world.config.user_data['plan_file'])
